# Copyright (C) Unitary Foundation
#
# This source code is licensed under the GPL license (v3) found in the
# LICENSE file in the root directory of this source tree.

"""Tests for zero-noise inference and extrapolation methods (factories) with
classically generated data.
"""

from collections.abc import Callable

import cirq
import numpy as np
from cirq import H, LineQubit, X
from numpy.random import RandomState
from pytest import mark, raises, warns

from mitiq.zne.inference import (
    AdaExpFactory,
    AdaptiveFactory,
    ConvergenceWarning,
    ExpFactory,
    ExtrapolationError,
    ExtrapolationWarning,
    FakeNodesFactory,
    LinearFactory,
    PolyExpFactory,
    PolyFactory,
    RichardsonFactory,
)

# Constant parameters for test functions:
A = 0.5
B = 0.7
C = 0.4
D = 0.3
X_VALS = [1, 1.3, 1.7, 2.2, 2.4]

# FakeNodesFactory only accepts equally spaced scale factors
UNIFORM_X = np.linspace(1.0, 3.0, 20)

# FakeNodesFactory needs a larger tolerance
LARGE_TOL = 1.0e-1

STAT_NOISE = 0.0001
CLOSE_TOL = 1.0e-2
# PolyExp fit is non-linear, so we set a larger tolerance
POLYEXP_TOL = 2 * CLOSE_TOL
NOT_CLOSE_TOL = 1.0e-1

# Set a seed.
SEED = 808


def apply_seed_to_func(func: Callable, seed: int) -> Callable:
    """Applies the input seed to the input function by
    using a random state and returns the seeded function."""
    rnd_state = RandomState(seed)

    def seeded_func(x: float, err: float = STAT_NOISE) -> float:
        return func(x, err=err, rnd_state=rnd_state)

    return seeded_func


# Classical test functions with statistical error:
def f_lin(
    x: float, err: float = STAT_NOISE, rnd_state: RandomState = np.random
) -> float:
    """Linear function."""
    return A + B * x + rnd_state.normal(scale=err)


def f_non_lin(
    x: float, err: float = STAT_NOISE, rnd_state: RandomState = np.random
) -> float:
    """Non-linear function."""
    return A + B * x + C * x**2 + rnd_state.normal(scale=err)


def f_exp_down(
    x: float, err: float = STAT_NOISE, rnd_state: RandomState = np.random
) -> float:
    """Exponential decay."""
    return A + B * np.exp(-C * x) + rnd_state.normal(scale=err)


def f_exp_up(
    x: float, err: float = STAT_NOISE, rnd_state: RandomState = np.random
) -> float:
    """Exponential growth."""
    return A - B * np.exp(-C * x) + rnd_state.normal(scale=err)


def f_poly_exp_down(
    x: float, err: float = STAT_NOISE, rnd_state: RandomState = np.random
) -> float:
    """Poly-exponential decay."""
    return A + B * np.exp(-C * x - D * x**2) + rnd_state.normal(scale=err)


def f_poly_exp_up(
    x: float, err: float = STAT_NOISE, rnd_state: RandomState = np.random
) -> float:
    """Poly-exponential growth."""
    return A - B * np.exp(-C * x - D * x**2) + rnd_state.normal(scale=err)


def f_lin_shot(x: float, shots=1) -> float:
    """Linear function with "shots" argument."""
    return A + B * x + 0.001 / np.sqrt(shots)


def f_runge(x: float) -> float:
    """Function with a peak at x=3, which has a strong Runge effect
    with RichardsonFactory but not with FakeNodesFactory.
    """
    return 1.0 / ((x - 2) ** 2 + 1.0)


@mark.parametrize("test_f", [f_lin, f_non_lin])
def test_noise_seeding(test_f: Callable[[float], float]):
    """Check that seeding works as expected."""
    seeded_f = apply_seed_to_func(test_f, SEED)
    noise_a = seeded_f(0)
    noise_b = seeded_f(0)
    seeded_f = apply_seed_to_func(test_f, SEED)
    noise_c = seeded_f(0)
    assert noise_a != noise_b
    assert noise_a == noise_c


@mark.parametrize(
    "factory",
    (
        LinearFactory,
        RichardsonFactory,
        FakeNodesFactory,
        PolyFactory,
        ExpFactory,
        PolyExpFactory,
    ),
)
def test_get_scale_factors_static_factories(factory):
    scale_factors = np.linspace(1.0, 10.0, num=20)
    if factory is PolyFactory or factory is PolyExpFactory:
        fac = factory(scale_factors=scale_factors, order=2)
    else:
        fac = factory(scale_factors=scale_factors)

    # Expectation values haven't been computed at any scale factors yet
    if isinstance(fac, AdaptiveFactory):
        assert not fac.get_scale_factors()
    else:
        assert isinstance(fac.get_scale_factors(), list)
        assert np.allclose(fac.get_scale_factors(), scale_factors)

    # Compute expectation values at all the scale factors
    fac.run_classical(apply_seed_to_func(f_lin, seed=1))
    assert isinstance(fac.get_scale_factors(), list)
    assert np.allclose(fac.get_scale_factors(), scale_factors)


@mark.parametrize("factory", (AdaExpFactory,))
def test_get_scale_factors_adaptive_factories(factory):
    num_steps = 8
    fac = AdaExpFactory(steps=num_steps, scale_factor=2.0, asymptote=None)

    # Expectation values haven't been computed at any scale factors yet
    assert isinstance(fac.get_scale_factors(), list)
    assert len(fac.get_scale_factors()) == 0

    # Compute expectation values at all the scale factors
    fac.run_classical(apply_seed_to_func(f_exp_up, seed=1))
    assert isinstance(fac.get_scale_factors(), list)

    # Given this seeded executor, the scale factors should be as follows
    correct_scale_factors = [
        1.0,
        2.0,
        4.0,
        4.20469548,
        4.20310693,
        4.2054822,
        4.2031916,
        4.2052843,
    ]
    assert len(fac.get_scale_factors()) == num_steps
    assert np.allclose(fac.get_scale_factors(), correct_scale_factors)


@mark.parametrize(
    "factory",
    (
        LinearFactory,
        RichardsonFactory,
        FakeNodesFactory,
        PolyFactory,
        ExpFactory,
        PolyExpFactory,
    ),
)
def test_get_expectation_values_static_factories(factory):
    scale_factors = np.linspace(1.0, 10.0, num=20)
    executor = apply_seed_to_func(f_lin, seed=1)
    expectation_values = [executor(scale) for scale in scale_factors]

    if factory is PolyFactory or factory is PolyExpFactory:
        fac = factory(scale_factors=scale_factors, order=2)
    else:
        fac = factory(scale_factors=scale_factors)

    # Expectation values haven't been computed at any scale factors yet
    assert isinstance(fac.get_expectation_values(), list)
    assert len(fac.get_expectation_values()) == 0

    # Compute expectation values at all the scale factors
    fac.run_classical(executor)
    assert isinstance(fac.get_expectation_values(), list)
    assert np.allclose(
        fac.get_expectation_values(), expectation_values, atol=1e-3
    )


@mark.parametrize("factory", (AdaExpFactory,))
def test_get_expectation_values_adaptive_factories(factory):
    num_steps = 8
    fac = AdaExpFactory(steps=num_steps, scale_factor=2.0, asymptote=None)
    executor = apply_seed_to_func(f_exp_up, seed=1)

    # Expectation values haven't been computed at any scale factors yet
    assert isinstance(fac.get_expectation_values(), list)
    assert len(fac.get_expectation_values()) == 0

    # Compute expectation values at all the scale factors
    fac.run_classical(executor)
    assert isinstance(fac.get_scale_factors(), list)

    # Given this seeded executor, the scale factors should be as follows
    correct_scale_factors = [
        1.0,
        2.0,
        4.0,
        4.20469548,
        4.20310693,
        4.2054822,
        4.2031916,
        4.2052843,
    ]
    correct_expectation_values = [
        executor(scale) for scale in correct_scale_factors
    ]
    assert len(fac.get_expectation_values()) == num_steps
    assert np.allclose(
        fac.get_expectation_values(), correct_expectation_values, atol=1e-3
    )


@mark.parametrize(
    "factory",
    (
        LinearFactory,
        RichardsonFactory,
        FakeNodesFactory,
        PolyFactory,
        ExpFactory,
        PolyExpFactory,
    ),
)
@mark.parametrize("batched", (True, False))
def test_run_sequential_and_batched(factory, batched):
    scale_factors = np.linspace(1.0, 10.0, num=20)

    if factory is PolyFactory or factory is PolyExpFactory:
        fac = factory(scale_factors=scale_factors, order=2)
    else:
        fac = factory(scale_factors=scale_factors)

    # Expectation values haven't been computed at any scale factors yet
    assert isinstance(fac.get_expectation_values(), list)
    assert len(fac.get_expectation_values()) == 0

    # Compute expectation values at all the scale factors
    if batched:

        def executor(circuits) -> list[float]:
            return [1.0] * len(circuits)

    else:

        def executor(circuit):
            return 1.0

    fac.run(cirq.Circuit(), executor, scale_noise=lambda circ, _: circ)

    assert isinstance(fac.get_expectation_values(), list)
    assert np.allclose(
        fac.get_expectation_values(), np.ones_like(scale_factors)
    )


@mark.parametrize("test_f", [f_lin, f_non_lin])
def test_richardson_extr(test_f: Callable[[float], float]):
    """Test of the Richardson's extrapolator."""
    seeded_f = apply_seed_to_func(test_f, SEED)
    fac = RichardsonFactory(scale_factors=X_VALS)
    assert not fac._opt_params
    fac.run_classical(seeded_f)
    zne_value = fac.reduce()
    assert np.isclose(zne_value, seeded_f(0, err=0), atol=CLOSE_TOL)
    assert len(fac._opt_params) == len(X_VALS)
    assert np.isclose(fac._opt_params[-1], zne_value)
    exp_vals = fac.get_expectation_values()
    assert np.isclose(fac.extrapolate(X_VALS, exp_vals), zne_value)
    assert np.isclose(
        fac.extrapolate(X_VALS, exp_vals, full_output=True)[0],
        zne_value,
    )


def test_fake_nodes_factory():
    """Test FakeNodesFactory in a specific regime in which the fake nodes
    interpolation method works well.
    """
    fac = FakeNodesFactory(UNIFORM_X)
    assert not fac._opt_params
    fac.run_classical(f_runge)
    zne_value = fac.reduce()
    assert np.isclose(zne_value, f_runge(0.0), atol=LARGE_TOL)
    assert len(fac._opt_params) == len(UNIFORM_X)
    assert np.isclose(fac._opt_params[-1], zne_value)
    assert np.isclose(
        fac.extrapolate(UNIFORM_X, f_runge(UNIFORM_X)), zne_value
    )
    exp_vals = fac.get_expectation_values()
    assert np.isclose(fac.extrapolate(UNIFORM_X, exp_vals), zne_value)
    assert np.isclose(
        fac.extrapolate(UNIFORM_X, exp_vals, full_output=True)[0],
        zne_value,
    )


def test_fake_nodes_extrapolation():
    """Test that there exists a regime in which FakeNodesFactory
    is better than RichardsonFactory.
    Note: in many cases RichardsonFactory is better.
    """
    y_vals = [f_runge(x) for x in UNIFORM_X]
    zne_runge = FakeNodesFactory.extrapolate(UNIFORM_X, y_vals)
    zne_richard = RichardsonFactory.extrapolate(UNIFORM_X, y_vals)
    abs_err_runge = np.abs(zne_runge - f_runge(0.0))
    abs_err_richard = np.abs(zne_richard - f_runge(0.0))
    # Test Richardson extrapolation error is much larger
    assert 500 * abs_err_runge < abs_err_richard


def test_linear_extr():
    """Tests extrapolation with a LinearFactory."""
    seeded_f = apply_seed_to_func(f_lin, SEED)
    fac = LinearFactory(X_VALS)
    assert not fac._opt_params
    fac.run_classical(seeded_f)
    zne_value = fac.reduce()
    assert np.isclose(zne_value, seeded_f(0, err=0), atol=CLOSE_TOL)
    assert np.allclose(fac._opt_params, [B, A], atol=CLOSE_TOL)
    exp_vals = fac.get_expectation_values()
    assert np.isclose(fac.extrapolate(X_VALS, exp_vals), zne_value)
    assert np.isclose(
        fac.extrapolate(X_VALS, exp_vals, full_output=True)[0],
        zne_value,
    )


def test_poly_extr():
    """Test of polynomial extrapolator."""
    # test (order=1)
    fac = PolyFactory(X_VALS, order=1)
    fac.run_classical(f_lin)
    assert np.isclose(fac.reduce(), f_lin(0, err=0), atol=CLOSE_TOL)
    # test that, for some non-linear functions,
    # order=1 is bad while order=2 is better.
    seeded_f = apply_seed_to_func(f_non_lin, SEED)
    fac = PolyFactory(X_VALS, order=1)
    fac.run_classical(seeded_f)
    assert not np.isclose(fac.reduce(), seeded_f(0, err=0), atol=NOT_CLOSE_TOL)
    seeded_f = apply_seed_to_func(f_non_lin, SEED)
    fac = PolyFactory(X_VALS, order=2)
    fac.run_classical(seeded_f)
    zne_value = fac.reduce()
    assert np.isclose(fac.reduce(), seeded_f(0, err=0), atol=CLOSE_TOL)
    exp_vals = fac.get_expectation_values()
    assert np.isclose(fac.extrapolate(X_VALS, exp_vals, order=2), zne_value)
    assert np.isclose(
        fac.extrapolate(X_VALS, exp_vals, order=2, full_output=True)[0],
        zne_value,
    )


@mark.parametrize("order", [2, 3, 4, 5])
def test_opt_params_poly_factory(order: int):
    """Tests that optimal parameters are stored after calling the reduce
    method.
    """
    fac = PolyFactory(scale_factors=np.linspace(1, 10, 10), order=order)
    assert not fac._opt_params
    fac.run_classical(apply_seed_to_func(f_non_lin, seed=SEED))
    zne_value = fac.reduce()
    assert len(fac._opt_params) == order + 1
    assert np.isclose(fac._opt_params[-1], zne_value)


@mark.parametrize("avoid_log", [False, True])
@mark.parametrize("test_f", [f_exp_down, f_exp_up])
def test_exp_factory_with_asympt(
    test_f: Callable[[float], float], avoid_log: bool
):
    """Test of exponential extrapolator."""
    seeded_f = apply_seed_to_func(test_f, SEED)
    fac = ExpFactory(X_VALS, asymptote=A, avoid_log=avoid_log)
    fac.run_classical(seeded_f)
    assert not fac._opt_params
    zne_value = fac.reduce()
    assert np.isclose(zne_value, seeded_f(0, err=0), atol=CLOSE_TOL)
    # There are three parameters in the exponential ansatz
    assert len(fac._opt_params) == 3
    exp_vals = fac.get_expectation_values()
    assert np.isclose(
        fac.extrapolate(X_VALS, exp_vals, asymptote=A, avoid_log=avoid_log),
        zne_value,
    )
    assert np.isclose(
        fac.extrapolate(
            X_VALS,
            exp_vals,
            asymptote=A,
            avoid_log=avoid_log,
            full_output=True,
        )[0],
        zne_value,
    )


def test_exp_factory_bad_asympt():
    with raises(ValueError, match="must be either a float or None"):
        ExpFactory(X_VALS, asymptote=1j)


@mark.parametrize("test_f", [f_exp_down, f_exp_up])
def test_exp_factory_no_asympt(test_f: Callable[[float], float]):
    """Test of exponential extrapolator."""
    seeded_f = apply_seed_to_func(test_f, SEED)
    fac = ExpFactory(X_VALS, asymptote=None)
    fac.run_classical(seeded_f)
    assert not fac._opt_params
    zne_value = fac.reduce()
    assert np.isclose(zne_value, seeded_f(0, err=0), atol=CLOSE_TOL)
    # There are three parameters in the exponential ansatz
    assert len(fac._opt_params) == 3
    exp_vals = fac.get_expectation_values()
    assert np.isclose(fac.extrapolate(X_VALS, exp_vals), zne_value)
    assert np.isclose(
        fac.extrapolate(X_VALS, exp_vals, full_output=True)[0],
        zne_value,
    )


@mark.parametrize("avoid_log", [False, True])
@mark.parametrize("test_f", [f_poly_exp_down, f_poly_exp_up])
def test_poly_exp_factory_with_asympt(
    test_f: Callable[[float], float], avoid_log: bool
):
    """Test of (almost) exponential extrapolator."""
    # test that, for a non-linear exponent,
    # order=1 is bad while order=2 is better.
    seeded_f = apply_seed_to_func(test_f, SEED)
    fac = PolyExpFactory(X_VALS, order=1, asymptote=A, avoid_log=avoid_log)
    fac.run_classical(seeded_f)
    assert not np.isclose(fac.reduce(), seeded_f(0, err=0), atol=NOT_CLOSE_TOL)
    seeded_f = apply_seed_to_func(test_f, SEED)
    fac = PolyExpFactory(X_VALS, order=2, asymptote=A, avoid_log=avoid_log)
    fac.run_classical(seeded_f)
    assert not fac._opt_params

    zne_value = fac.reduce()
    assert np.isclose(zne_value, seeded_f(0, err=0), atol=POLYEXP_TOL)

    # There are four parameters to fit for the PolyExpFactory of order 1
    assert len(fac._opt_params) == 4

    exp_values = [test_f(x) for x in X_VALS]
    assert np.isclose(
        PolyExpFactory.extrapolate(
            X_VALS, exp_values, order=2, asymptote=A, avoid_log=avoid_log
        ),
        zne_value,
        atol=POLYEXP_TOL,
    )


@mark.parametrize("test_f", [f_poly_exp_down, f_poly_exp_up])
def test_poly_exp_factory_no_asympt(test_f: Callable[[float], float]):
    """Test of (almost) exponential extrapolator."""
    seeded_f = apply_seed_to_func(test_f, SEED)
    # test that, for a non-linear exponent,
    # order=1 is bad while order=2 is better.
    fac = PolyExpFactory(X_VALS, order=1, asymptote=None)
    fac.run_classical(seeded_f)
    assert not np.isclose(fac.reduce(), seeded_f(0, err=0), atol=NOT_CLOSE_TOL)
    seeded_f = apply_seed_to_func(test_f, SEED)
    fac = PolyExpFactory(X_VALS, order=2, asymptote=None)
    fac.run_classical(seeded_f)
    zne_value = fac.reduce()
    assert np.isclose(zne_value, seeded_f(0, err=0), atol=POLYEXP_TOL)

    exp_vals = fac.get_expectation_values()
    assert np.isclose(fac.extrapolate(X_VALS, exp_vals, order=2), zne_value)
    assert np.isclose(
        fac.extrapolate(
            X_VALS,
            exp_vals,
            order=2,
            full_output=True,
        )[0],
        zne_value,
    )


@mark.parametrize("infinite_noise_limit", [0.0, 0.2, 0.4, 0.6, 0.8, 1.0, 1.2])
@mark.parametrize(
    "exp_vals", [[0.700000001, 0.7, 0.7, 0.7], [0.7, 0.7, 0.7, 0.700001]]
)
def test_poly_exp_factory_converges_toward_asympt(
    exp_vals: list[float],
    infinite_noise_limit: float,
):
    """Test of (almost) exponential extrapolator in special cases"""
    # show that the extrapolation achieves the best fit even when the measured
    # data decay or grow away from the expected direction as indicated by the
    # asymptote
    scale_factors = [1.0, 2, 3, 4]
    assert np.isclose(
        ExpFactory.extrapolate(
            scale_factors, exp_vals, asymptote=infinite_noise_limit
        ),
        0.7,
    )


@mark.parametrize("avoid_log", [False, True])
@mark.parametrize("test_f", [f_exp_down, f_exp_up])
def test_ada_exp_factory_with_asympt(
    test_f: Callable[[float], float], avoid_log: bool
):
    """Test of the adaptive exponential extrapolator."""
    seeded_f = apply_seed_to_func(test_f, SEED)
    fac = AdaExpFactory(
        steps=3, scale_factor=2.0, asymptote=A, avoid_log=avoid_log
    )
    # Note: run_classical calls next which calls reduce, so calling
    # fac.run_classical with an AdaExpFactory sets the optimal parameters as
    # well. Hence we check that the opt_params are empty before
    # AdaExpFactory.run_classical is called.
    assert not fac._opt_params
    fac.run_classical(seeded_f)
    assert np.isclose(fac.reduce(), seeded_f(0, err=0), atol=CLOSE_TOL)

    # There are three parameters to fit for the (adaptive) exponential ansatz
    assert len(fac._opt_params) == 3


@mark.parametrize("avoid_log", [False, True])
@mark.parametrize("test_f", [f_exp_down, f_exp_up])
def test_ada_exp_fac_with_asympt_more_steps(
    test_f: Callable[[float], float], avoid_log: bool
):
    """Test of the adaptive exponential extrapolator with more steps."""
    seeded_f = apply_seed_to_func(test_f, SEED)
    fac = AdaExpFactory(
        steps=6, scale_factor=2.0, asymptote=A, avoid_log=avoid_log
    )
    fac.run_classical(seeded_f)
    assert np.isclose(fac.reduce(), seeded_f(0, err=0), atol=CLOSE_TOL)


@mark.parametrize("test_f", [f_exp_down, f_exp_up])
def test_ada_exp_factory_no_asympt(test_f: Callable[[float], float]):
    """Test of the adaptive exponential extrapolator."""
    seeded_f = apply_seed_to_func(test_f, SEED)
    fac = AdaExpFactory(steps=4, scale_factor=2.0, asymptote=None)
    fac.run_classical(seeded_f)
    zne_value = fac.reduce()
    assert np.isclose(zne_value, seeded_f(0, err=0), atol=CLOSE_TOL)


@mark.parametrize("test_f", [f_exp_down, f_exp_up])
def test_ada_exp_factory_no_asympt_more_steps(
    test_f: Callable[[float], float],
):
    """Test of the adaptive exponential extrapolator."""
    seeded_f = apply_seed_to_func(test_f, SEED)
    fac = AdaExpFactory(steps=8, scale_factor=2.0, asymptote=None)
    fac.run_classical(seeded_f)
    assert np.isclose(fac.reduce(), seeded_f(0, err=0), atol=CLOSE_TOL)


def test_ada_exp_factory_bad_arguments():
    with raises(ValueError, match="must be an integer greater or equal to 3"):
        AdaExpFactory(steps=2.5)

    with raises(ValueError, match="must be strictly larger than one"):
        AdaExpFactory(steps=4, scale_factor=0.5)

    with raises(ValueError, match="must be strictly larger than one"):
        AdaExpFactory(steps=4, max_scale_factor=1)

    with raises(ValueError, match="must be either a float or None"):
        AdaExpFactory(steps=10, asymptote=1j)


def test_avoid_log_keyword():
    """Test that avoid_log=True and avoid_log=False give different results."""
    fac = ExpFactory(X_VALS, asymptote=A, avoid_log=False)
    fac.run_classical(f_exp_down)
    znl_with_log = fac.reduce()
    fac._options["avoid_log"] = True
    znl_without_log = fac.reduce()
    assert not znl_with_log == znl_without_log


@mark.parametrize(
    "factory", (LinearFactory, RichardsonFactory, FakeNodesFactory)
)
def test_too_few_scale_factors(factory):
    """Test less than 2 scale_factors."""
    with raises(ValueError, match=r"At least 2 scale factors are necessary"):
        _ = factory([1])


def test_order_is_too_high_for_scale_factors():
    """Test that a wrong initialization error is raised."""
    with raises(ValueError, match=r"The extrapolation order cannot exceed"):
        _ = PolyFactory(X_VALS, order=10)


def test_too_few_points_for_polyfit_warning():
    """Test that the correct warning is raised if data is not enough to fit."""
    fac = PolyFactory(X_VALS, order=2)
    fac._instack = [
        {"scale_factor": 1.0, "shots": 100},
        {"scale_factor": 2.0, "shots": 100},
    ]
    fac._outstack = [1.0, 2.0]
    with warns(
        ExtrapolationWarning,
        match=r"The extrapolation fit may be ill-conditioned.",
    ):
        fac.reduce()
    # test also the static "extrapolate" method.
    with warns(
        ExtrapolationWarning,
        match=r"The extrapolation fit may be ill-conditioned.",
    ):
        PolyFactory.extrapolate([1.0, 2.0], [1.0, 2.0], order=2)


def test_failing_fit_error():
    """Test error handling for a failing fit."""
    fac = ExpFactory(X_VALS, asymptote=None)
    fac._instack = [{"scale_factor": x} for x in X_VALS]
    fac._outstack = [1.0, 2.0, 1.0, 2.0, 1.0]
    with raises(
        ExtrapolationError, match=r"The extrapolation fit failed to converge."
    ):
        fac.reduce()
    # test also the static "extrapolate" method.
    with raises(
        ExtrapolationError, match=r"The extrapolation fit failed to converge."
    ):
        ExpFactory.extrapolate(X_VALS, [1.0, 2.0, 1.0, 2.0, 1.0])


@mark.parametrize("fac", [LinearFactory([1, 1, 1]), ExpFactory([1, 1, 1])])
def test_failing_fit_warnings(fac):
    """Test that the correct warning is raised for an ill-conditioned fit."""
    fac._instack = [{"scale_factor": 1.0} for _ in range(4)]
    fac._outstack = [1, 1, 1, 1]
    with warns(
        ExtrapolationWarning,
        match=r"The extrapolation fit may be ill-conditioned.",
    ):
        fac.reduce()
    # test also the static "extrapolate" method.
    with warns(
        ExtrapolationWarning,
        match=r"The extrapolation fit may be ill-conditioned.",
    ):
        fac.extrapolate([1, 1, 1, 1], [1.0, 1.0, 1.0, 1.0])


def test_adaptive_factory_max_iteration_warnings():
    """Test that the correct warning is raised beyond the iteration limit."""
    fac = AdaExpFactory(steps=10)
    with warns(
        ConvergenceWarning,
        match=r"Factory iteration loop stopped before convergence.",
    ):
        fac.run_classical(lambda scale_factor: 1.0, max_iterations=3)


@mark.parametrize("fac_class", [LinearFactory, RichardsonFactory])
def test_iterate_with_shot_list(fac_class):
    """Tests factories with (and without) the "shot_list" argument."""
    # first test without shot_list
    fac = fac_class(X_VALS)
    fac.run_classical(f_lin_shot)
    assert np.isclose(fac.reduce(), f_lin_shot(0), atol=CLOSE_TOL)

    # Check instack and outstack are as expected
    SHOT_LIST = [100, 200, 300, 400, 500]
    for j, shots in enumerate(SHOT_LIST):
        assert fac._instack[j] == {"scale_factor": X_VALS[j]}
        assert fac._outstack[j] != f_lin_shot(X_VALS[j], shots=shots)
        assert fac._outstack[j] == f_lin_shot(X_VALS[j])

    # Now pass an arbitrary shot_list as an argument
    fac = fac_class(X_VALS, shot_list=SHOT_LIST)
    fac.run_classical(f_lin_shot)
    assert np.isclose(fac.reduce(), f_lin_shot(0), atol=CLOSE_TOL)

    # Check instack and outstack are as expected
    for j, shots in enumerate(SHOT_LIST):
        assert fac._instack[j] == {"scale_factor": X_VALS[j], "shots": shots}
        assert fac._outstack[j] == f_lin_shot(X_VALS[j], shots=shots)
        assert fac._outstack[j] != f_lin_shot(X_VALS[j])


def test_shot_list_errors():
    """Tests errors related to the "shot_lists" argument."""
    with raises(IndexError, match=r"must have the same length."):
        PolyFactory(X_VALS, order=2, shot_list=[1, 2])
    with raises(TypeError, match=r"valid iterator of integers"):
        PolyFactory(X_VALS, order=2, shot_list=[1.0, 2])


def test_push_after_already_reduced_warning():
    """Tests a warning is raised if new data is pushed in a factory
    which was already reduced."""
    fac = LinearFactory([1, 2])
    fac.push({"scale_factor": 1.0}, 1.0)
    fac.push({"scale_factor": 2.0}, 2.0)
    fac.reduce()
    with warns(
        ExtrapolationWarning,
        match=r"You are pushing new data into a factory object",
    ):
        fac.push({"scale_factor": 3.0}, 3.0)
    # Assert no warning is raised when .reset() is used
    fac.reset()
    fac.push({"scale_factor": 1.0}, 2.0)
    fac.push({"scale_factor": 2.0}, 1.0)
    assert np.isclose(3.0, fac.reduce())


def test_full_output_keyword():
    """Tests the full_output keyword in extrapolate method."""
    zne_limit = LinearFactory.extrapolate([1, 2], [1, 2])
    assert np.isclose(zne_limit, 0.0)
    (
        zne_limit,
        zne_std,
        opt_params,
        params_cov,
        zne_curve,
    ) = LinearFactory.extrapolate([1, 2], [1, 2], full_output=True)

    assert len(opt_params) == 2
    assert np.isclose(zne_limit, 0.0)
    assert np.isclose(0.0, opt_params[1])
    assert np.isclose(1.0, opt_params[0])
    assert zne_std is None
    assert params_cov is None
    assert np.isclose(zne_curve(0), 0.0)
    assert np.isclose(zne_curve(2), 2.0)


def test_full_output_keyword_cov_std():
    """Tests the full_output keyword in extrapolate method."""
    zne_limit = PolyFactory.extrapolate([1, 2, 3], [1, 4, 9], order=2)
    assert np.isclose(zne_limit, 0.0)
    (
        zne_limit,
        zne_std,
        opt_params,
        params_cov,
        zne_curve,
    ) = PolyFactory.extrapolate(
        [1, 2, 3], [1, 4, 9], order=2, full_output=True
    )

    assert len(opt_params) == 3
    assert np.isclose(zne_limit, 0.0)
    assert np.isclose(0.0, opt_params[1])
    assert np.isclose(1.0, opt_params[0])
    assert params_cov is None
    assert zne_std is None
    assert np.isclose(zne_curve(0), 0.0)
    assert np.isclose(zne_curve(2), 4.0)
    assert np.isclose(zne_curve(3), 9.0)


def test_params_cov_and_zne_std():
    """Tests the variance of the parametes and of the zne are produced."""
    x_values = [0, 0, 1]
    y_values = [-1, 1, 0]
    zne_limit = PolyFactory.extrapolate(x_values, y_values, order=1)
    assert np.isclose(zne_limit, 0.0, atol=1.0e-4)
    (
        zne_limit,
        zne_std,
        opt_params,
        params_cov,
        zne_curve,
    ) = PolyFactory.extrapolate(x_values, y_values, order=1, full_output=True)
    assert len(opt_params) == 2
    assert np.isclose(zne_limit, 0.0)
    assert np.isclose(0.0, opt_params[1])
    assert np.isclose(0.0, opt_params[0])
    assert np.allclose(params_cov, [[3.0, -1.0], [-1.0, 1.0]])
    assert np.isclose(zne_std, 1.0)
    assert np.isclose(zne_curve(0), 0.0)
    assert np.isclose(zne_curve(0.5), 0.0)


@mark.parametrize(
    "factory", [LinearFactory, RichardsonFactory, FakeNodesFactory]
)
def test_execute_with_zne_fit_fail(factory):
    """Tests errors are raised when asking for fitting parameters that can't
    be calculated.
    """
    with raises(ValueError, match="Data is either ill-defined or not enough"):
        factory([1.0, 2.0]).get_zero_noise_limit_error()
    with raises(ValueError, match="Data is either ill-defined or not enough"):
        factory([1.0, 2.0]).get_optimal_parameters()
    with raises(ValueError, match="Data is either ill-defined or not enough"):
        factory([1.0, 2.0]).get_parameters_covariance()
    with raises(ValueError, match="Data is either ill-defined or not enough"):
        factory([1.0, 2.0]).get_zero_noise_limit()
    with raises(ValueError, match="Data is either ill-defined or not enough"):
        factory([1.0, 2.0]).get_extrapolation_curve()


def test_get_methods_of_factories():
    """Tests the get methods of a factory"""
    x_values = [0, 0, 1]
    y_values = [-1, 1, 0]
    fac = LinearFactory(x_values)
    fac._instack = [
        {"scale_factor": 0},
        {"scale_factor": 0},
        {"scale_factor": 1},
    ]
    fac._outstack = y_values
    zne_reduce = fac.reduce()

    assert np.allclose(fac.get_expectation_values(), y_values)
    assert np.allclose(fac.get_extrapolation_curve()(0.0), zne_reduce)
    assert np.allclose(fac.get_optimal_parameters(), [0.0, 0.0])
    assert np.allclose(
        fac.get_parameters_covariance(), [[3.0, -1.0], [-1.0, 1.0]]
    )
    assert np.allclose(fac.get_scale_factors(), x_values)
    assert np.allclose(fac.get_zero_noise_limit(), zne_reduce)
    assert np.allclose(fac.get_zero_noise_limit_error(), 1.0)


@mark.parametrize(
    "factory",
    (
        LinearFactory,
        RichardsonFactory,
        PolyFactory,
    ),
)
def test_plot_data(factory):
    """Test that plot_data() outputs the correct x and y values."""
    if factory is PolyFactory:
        fac = factory(scale_factors=X_VALS, order=3)
    else:
        fac = factory(scale_factors=X_VALS)
    fac.run_classical(apply_seed_to_func(f_lin, SEED))
    fac.reduce()
    fig = fac.plot_data()
    ax = fig.gca()
    x_data, y_data = ax.lines[0].get_xydata().T
    zne_curve = fac.get_extrapolation_curve()
    np.allclose(y_data, zne_curve(x_data))


@mark.parametrize("factory", (ExpFactory, PolyExpFactory))
def test_plot_data_exp_factory(factory):
    """Test that plot_data() outputs the correct x and y values
    for ExpFactory and PolyExpFactory.
    """
    if factory is PolyExpFactory:
        fac = factory(scale_factors=X_VALS, order=2)
    else:
        fac = factory(scale_factors=X_VALS)
    fac.run_classical(apply_seed_to_func(f_exp_down, SEED))
    fac.reduce()
    fig = fac.plot_data()
    ax = fig.gca()
    x_data, y_data = ax.lines[0].get_xydata().T
    zne_curve = fac.get_extrapolation_curve()
    np.allclose(y_data, zne_curve(x_data))


@mark.parametrize(
    "factory",
    (
        LinearFactory,
        RichardsonFactory,
        PolyFactory,
    ),
)
def test_plot_fit(factory):
    """Test that plot_fit() outputs the correct x and y values."""
    if factory is PolyFactory or factory is PolyExpFactory:
        fac = factory(scale_factors=X_VALS, order=2)
    else:
        fac = factory(scale_factors=X_VALS)
    fac.run_classical(apply_seed_to_func(f_lin, SEED))
    fac.reduce()
    fig = fac.plot_fit()
    ax = fig.gca()
    x_data, y_data = ax.lines[0].get_xydata().T
    zne_curve = fac.get_extrapolation_curve()
    np.allclose(y_data, zne_curve(x_data))


@mark.parametrize("factory", (ExpFactory, PolyExpFactory))
def test_plot_fit_exp_factory(factory):
    """Test that plot_fit() outputs the correct x and y values
    for ExpFactory and PolyExpFactory.
    """
    if factory is PolyExpFactory:
        fac = factory(scale_factors=X_VALS, order=2)
    else:
        fac = factory(scale_factors=X_VALS)
    fac.run_classical(apply_seed_to_func(f_exp_down, SEED))
    fac.reduce()
    fig = fac.plot_fit()
    ax = fig.gca()
    x_data, y_data = ax.lines[0].get_xydata().T
    zne_curve = fac.get_extrapolation_curve()
    np.allclose(y_data, zne_curve(x_data))


def test_fakenodes_scale_factors_equally_spaced():
    """FakeNodesFactory should only accept equally spaced scale factors."""
    y_vals = [0.5, 1.0, 1.5]
    with raises(
        ValueError, match=r"The scale factors must be equally spaced."
    ):
        _ = FakeNodesFactory(X_VALS).extrapolate(X_VALS, y_vals)


def test_map_to_fakenodes():
    """Test the fake nodes map in FakeNodesFactory."""
    fac = FakeNodesFactory(UNIFORM_X)
    test_argument = 1.0
    assert np.isclose(
        fac._map_to_fake_nodes([1.0], 2.0, test_argument),
        [1.0],
    )


@mark.parametrize(
    "factory",
    (
        LinearFactory,
        RichardsonFactory,
        FakeNodesFactory,
        PolyFactory,
        ExpFactory,
        PolyExpFactory,
        AdaExpFactory,
    ),
)
def test_short_circuit_warning(factory):
    """Tests a warning is raised if the input circuit has very few gates."""
    scale_factors = np.linspace(1.0, 10.0, num=20)

    def executor(circuits) -> list[float]:
        return [1.0] * len(circuits)

    if factory is PolyFactory or factory is PolyExpFactory:
        fac = factory(scale_factors=scale_factors, order=2)
    elif factory is AdaExpFactory:
        fac = factory(steps=4)
    else:
        fac = factory(scale_factors=scale_factors)

    qubit = LineQubit(0)
    circuit = cirq.Circuit(X(qubit), H(qubit), X(qubit))

    with warns(
        UserWarning,
        match=r"The input circuit is very short.",
    ):
        fac.run(circuit, executor, scale_noise=lambda circ, _: circ)


@mark.xfail(reason="cannot extrapolate complex values")
def test_extrapolate_raises_error_on_complex_expectation_values():
    PolyExpFactory.extrapolate(
        scale_factors=[1, 2, 3], exp_values=[1, 0.5 + 1e-5j, 6], order=1
    )



# Copyright (C) Unitary Foundation
#
# This source code is licensed under the GPL license (v3) found in the
# LICENSE file in the root directory of this source tree.

"""Tests for mitiq.about()."""

import mitiq


def test_result_and_stdout(capsys):
    mitiq.about()
    captured = capsys.readouterr()
    assert captured.out.startswith(
        "\nMitiq: A Python toolkit for implementing"
    )



# Copyright (C) Unitary Foundation
#
# This source code is licensed under the GPL license (v3) found in the
# LICENSE file in the root directory of this source tree.

"""Test classical shadow estimation process."""

from numbers import Number

import cirq

import mitiq
from mitiq import MeasurementResult
from mitiq.interface.mitiq_cirq.cirq_utils import (
    sample_bitstrings as cirq_sample_bitstrings,
)
from mitiq.shadows.shadows import (
    classical_post_processing,
    pauli_twirling_calibrate,
    shadow_quantum_processing,
)

# define a fully entangled state
num_qubits: int = 2
qubits = cirq.LineQubit.range(num_qubits)
circuit = cirq.Circuit([cirq.H(q) for q in qubits])
circuit.append(cirq.CNOT(qubits[0], qubits[1]))
observables = [mitiq.PauliString("X", support=(i,)) for i in range(num_qubits)]


def executor(
    circuit: cirq.Circuit,
) -> MeasurementResult:
    return cirq_sample_bitstrings(
        circuit,
        noise_level=(0,),
        shots=1,
        sampler=cirq.Simulator(),
    )


def test_pauli_twirling_calibrate():
    # Call the function with valid inputs
    result = pauli_twirling_calibrate(
        qubits=qubits, executor=executor, num_total_measurements_calibration=2
    )

    # Check that the dictionary contains the correct number of entries
    assert len(result) <= 2**num_qubits

    for value in result.values():
        assert isinstance(value, Number)

    # Call shadow_quantum_processing to get shadow_outcomes
    shadow_outcomes = (["11", "00"], ["ZZ", "XX"])

    # Call the function with valid inputs
    result = pauli_twirling_calibrate(
        zero_state_shadow_outcomes=shadow_outcomes,
        num_total_measurements_calibration=2,
    )

    # Check that the dictionary contains the correct number of entries
    assert len(result) <= 2**num_qubits

    for value in result.values():
        assert isinstance(value, Number)


def test_shadow_quantum_processing():
    # Call the function with valid inputs
    result = shadow_quantum_processing(
        circuit, executor, num_total_measurements_shadow=10
    )

    # Check that the result is a tuple
    assert isinstance(result, tuple), f"Expected a tuple, got {type(result)}"

    # Check that the tuple contains two lists
    assert (
        len(result) == 2
    ), f"Expected two lists in the tuple, got {len(result)}"
    assert isinstance(result[0], list)
    assert isinstance(result[1], list)


def test_classical_post_processing():
    # Call shadow_quantum_processing to get shadow_outcomes
    shadow_outcomes = (["11", "00"], ["ZZ", "XX"])

    # Call pauli_twirling_calibrate to get calibration_results
    calibration_results = {"00": 1, "01": 1 / 3, "10": 1 / 3, "11": 1 / 9}

    # Call the function with valid inputs and state_reconstruction=True
    result = classical_post_processing(
        shadow_outcomes, state_reconstruction=True
    )

    # Check that the result is a dictionary
    assert isinstance(
        result, dict
    ), f"Expected a dictionary, got {type(result)}"

    # Check that the dictionary contains the expected keys
    assert "reconstructed_state" in result

    # Call the function with valid inputs and observables provided
    result = classical_post_processing(
        shadow_outcomes, observables=observables
    )
    result_cal = classical_post_processing(
        shadow_outcomes,
        calibration_results=calibration_results,
        observables=observables,
        k_shadows=1,
    )
    # Check that the result is a dictionary
    assert isinstance(result, dict)
    assert result_cal == result

    # Check that the dictionary contains the expected keys
    for obs in observables:
        assert str(obs) in result
        assert isinstance(result[str(obs)], float)



# Copyright (C) Unitary Foundation
#
# This source code is licensed under the GPL license (v3) found in the
# LICENSE file in the root directory of this source tree.

"""Tests for GHZ circuits."""

import numpy as np
import pytest

from mitiq import SUPPORTED_PROGRAM_TYPES
from mitiq.benchmarks import ghz_circuits


@pytest.mark.parametrize("nqubits", [1, 5])
def test_ghz_circuits(nqubits):
    # test GHZ creation
    circuit = ghz_circuits.generate_ghz_circuit(nqubits)
    # check that the state |0...0> and the state |1...1>
    # both have probability close to 0.5
    sv = circuit.final_state_vector()
    zero_prob = abs(sv[0]) ** 2
    one_prob = abs(sv[-1]) ** 2
    assert np.isclose(zero_prob, 0.5)
    assert np.isclose(one_prob, 0.5)


def test_ghz_value_error():
    # test GHZ error raising
    with pytest.raises(ValueError):
        circuit = ghz_circuits.generate_ghz_circuit(0)  # noqa: F841


@pytest.mark.parametrize("return_type", SUPPORTED_PROGRAM_TYPES.keys())
def test_ghz_conversion(return_type):
    circuit = ghz_circuits.generate_ghz_circuit(3, return_type)
    assert return_type in circuit.__module__



# Copyright (C) Unitary Foundation
#
# This source code is licensed under the GPL license (v3) found in the
# LICENSE file in the root directory of this source tree.

import cirq
import numpy as np
import pytest

from mitiq.vd.vd_utils import (
    _apply_cyclic_system_permutation,
    _apply_diagonalizing_gate,
    _apply_symmetric_observable,
    _copy_circuit_parallel,
    _generate_diagonalizing_gate,
)


def test_copy_circuit_parallel_lengths():
    circuit = cirq.Circuit(
        cirq.H(cirq.LineQubit(0)),
        cirq.SWAP(cirq.LineQubit(0), cirq.LineQubit(1)),
    )
    for num in range(2, 10):
        new_circuit = _copy_circuit_parallel(circuit, num)
        assert len(new_circuit.all_qubits()) == 2 * num

    circuit = cirq.Circuit(
        cirq.X(cirq.LineQubit(0)),
        cirq.Y(cirq.LineQubit(1)),
        cirq.Z(cirq.LineQubit(2)),
    )
    for num in range(2, 10):
        new_circuit = _copy_circuit_parallel(circuit, num)
        assert len(new_circuit.all_qubits()) == 3 * num


def test_copy_circuit_parallel_full():
    num = 2
    qubits = cirq.LineQubit.range(2)
    circuit = cirq.Circuit(cirq.H(qubits[0]), cirq.CNOT(qubits[0], qubits[1]))
    new_circuit = _copy_circuit_parallel(circuit, num)

    expected_qubits = cirq.LineQubit.range(2 * num)
    expected_circuit = cirq.Circuit(
        cirq.H(expected_qubits[0]),
        cirq.CNOT(expected_qubits[0], expected_qubits[1]),
        cirq.H(expected_qubits[2]),
        cirq.CNOT(expected_qubits[2], expected_qubits[3]),
    )

    new_unitary = cirq.unitary(new_circuit)
    expected_unitary = cirq.unitary(expected_circuit)
    assert np.allclose(new_unitary, expected_unitary)

    num = 3
    qubits = cirq.LineQubit.range(1)
    circuit = cirq.Circuit(cirq.X(qubits[0]))
    new_circuit = _copy_circuit_parallel(circuit, num)

    expected_qubits = cirq.LineQubit.range(1 * num)
    expected_circuit = cirq.Circuit(
        cirq.X(expected_qubits[0]),
        cirq.X(expected_qubits[1]),
        cirq.X(expected_qubits[2]),
    )

    new_unitary = cirq.unitary(new_circuit)
    expected_unitary = cirq.unitary(expected_circuit)
    assert np.allclose(new_unitary, expected_unitary)


def test_copy_circuit_parallel_gridqubits():
    num = 3
    qubits = cirq.GridQubit.rect(2, 2)
    circuit = cirq.Circuit(
        cirq.H(qubits[0]),
        cirq.CNOT(qubits[0], qubits[1]),
        cirq.CNOT(qubits[0], qubits[2]),
    )

    new_circuit = _copy_circuit_parallel(circuit, num)
    new_qubits = new_circuit.all_qubits()

    assert len(new_qubits) == len(circuit.all_qubits()) * num

    expected_qubits = cirq.GridQubit.rect(2 * num, 2)
    expected_circuit = cirq.Circuit(
        cirq.H(expected_qubits[0]),
        cirq.CNOT(expected_qubits[0], expected_qubits[1]),
        cirq.CNOT(expected_qubits[0], expected_qubits[2]),
        cirq.H(expected_qubits[4]),
        cirq.CNOT(expected_qubits[4], expected_qubits[5]),
        cirq.CNOT(expected_qubits[4], expected_qubits[6]),
        cirq.H(expected_qubits[8]),
        cirq.CNOT(expected_qubits[8], expected_qubits[9]),
        cirq.CNOT(expected_qubits[8], expected_qubits[10]),
    )

    new_unitary = cirq.unitary(new_circuit)
    expected_unitary = cirq.unitary(expected_circuit)
    assert np.allclose(new_unitary, expected_unitary)


def test_apply_diagonalizing_gate_keeps_circuit_structure():
    num_copies = 2
    qubits = cirq.LineQubit.range(3)
    circuit = cirq.Circuit(
        cirq.H(qubits[0]),
        cirq.CNOT(qubits[0], qubits[1]),
        cirq.X(qubits[2]),
    )
    num_qubits = len(circuit.all_qubits())

    copied_circuit = _copy_circuit_parallel(circuit, num_copies)
    new_circuit = _apply_diagonalizing_gate(copied_circuit, num_copies)

    num_ops = len(list(copied_circuit.all_operations()))
    copied_circuit_ops = list(new_circuit.all_operations())

    assert len(copied_circuit_ops) == num_ops + num_qubits
    assert set(new_circuit.all_qubits()) == set(copied_circuit.all_qubits())


@pytest.mark.xfail(reason="VD does not yet support grid qubits")
def test_apply_diagonalizing_gate_fails_on_grid_qubits():
    num_copies = 2
    qubits = cirq.GridQubit.rect(2, 2)
    circuit = cirq.Circuit(
        cirq.H(qubits[0]),
        cirq.CNOT(qubits[0], qubits[1]),
        cirq.X(qubits[2]),
    )
    num_qubits = len(circuit.all_qubits())

    copied_circuit = _copy_circuit_parallel(circuit, num_copies)
    new_circuit = _apply_diagonalizing_gate(copied_circuit, num_copies)

    num_ops = len(list(copied_circuit.all_operations()))
    copied_circuit_ops = list(new_circuit.all_operations())

    assert len(copied_circuit_ops) == num_ops + num_qubits  # this passes
    assert set(new_circuit.all_qubits()) == set(copied_circuit.all_qubits())


def test_generate_diagonalizing_gate():
    expected_matrix = np.array(
        [
            [1, 0, 0, 0],
            [0, np.sqrt(2) / 2, np.sqrt(2) / 2, 0],
            [0, np.sqrt(2) / 2, -np.sqrt(2) / 2, 0],
            [0, 0, 0, 1],
        ]
    )

    assert np.allclose(
        _generate_diagonalizing_gate(2)._matrix, expected_matrix
    )


def test_apply_cyclic_system_permutation():
    # Check for ndim=2
    matrix = np.arange(1, 17).reshape(4, 4)
    expected = matrix[[0, 2, 1, 3]]

    result = _apply_cyclic_system_permutation(matrix, N_qubits=1)
    assert np.allclose(result, expected)

    # Check for ndim=3
    tensor = np.arange(1, 65).reshape(4, 4, 4)
    expected = tensor[:, [0, 2, 1, 3], :]

    result = _apply_cyclic_system_permutation(tensor, N_qubits=1)
    assert np.allclose(result, expected)

    # Check for ndim>3
    tensor = np.arange(1, 17).reshape(2, 2, 2, 2)

    with pytest.raises(TypeError, match="matrix should be 2D or 3D ndarray"):
        _apply_cyclic_system_permutation(tensor, N_qubits=1)


def test_edge_case_single_qubit():
    matrix = np.array([[1, 2], [3, 4]])
    # No permutation should happen here
    expected = matrix.copy()
    result = _apply_cyclic_system_permutation(
        matrix, N_qubits=1, num_registers=1
    )
    assert np.allclose(result, expected)


def test_apply_symmetric_observable():
    # Check for ndim=2
    matrix = np.arange(1, 17).reshape(4, 4)
    expected = np.array(
        [
            [
                [1.0, 0.0, 0.0, -4.0],
                [5.0, 0.0, 0.0, -8.0],
                [9.0, 0.0, 0.0, -12.0],
                [13.0, 0.0, 0.0, -16.0],
            ]
        ]
    )

    result = _apply_symmetric_observable(matrix, N_qubits=1)
    assert np.allclose(result, expected)

    # Check for ndim=3
    tensor = np.arange(1, 65).reshape(4, 4, 4)
    expected = np.array(
        [
            [
                [1.0, 0.0, 0.0, -4.0],
                [5.0, 0.0, 0.0, -8.0],
                [9.0, 0.0, 0.0, -12.0],
                [13.0, 0.0, 0.0, -16.0],
            ],
            [
                [17.0, 0.0, 0.0, -20.0],
                [21.0, 0.0, 0.0, -24.0],
                [25.0, 0.0, 0.0, -28.0],
                [29.0, 0.0, 0.0, -32.0],
            ],
            [
                [33.0, 0.0, 0.0, -36.0],
                [37.0, 0.0, 0.0, -40.0],
                [41.0, 0.0, 0.0, -44.0],
                [45.0, 0.0, 0.0, -48.0],
            ],
            [
                [49.0, 0.0, 0.0, -52.0],
                [53.0, 0.0, 0.0, -56.0],
                [57.0, 0.0, 0.0, -60.0],
                [61.0, 0.0, 0.0, -64.0],
            ],
        ]
    )

    result = _apply_symmetric_observable(tensor, N_qubits=1)
    assert np.allclose(result, expected)

    # Check for ndim>3
    tensor = np.arange(1, 17).reshape(2, 2, 2, 2)

    with pytest.raises(ValueError, match="matrix should be 2D or 3D ndarray"):
        _apply_symmetric_observable(tensor, N_qubits=1)


def test_apply_symmetric_observable_X():
    # For N_qubits=1 and M=2, we're dealing with 4x4 matrices
    matrix = np.arange(1, 17).reshape(4, 4)
    observable = np.array([[0, 1], [1, 0]])

    # For N_qubits=1, we expect one output matrix with diagonals scaled
    # by +1 for |0⟩ state and -1 for |1⟩ state
    expected = np.array(
        [
            [
                [7.0, 8.0, 9.0, 10.0],
                [7.0, 8.0, 9.0, 10.0],
                [7.0, 8.0, 9.0, 10.0],
                [7.0, 8.0, 9.0, 10.0],
            ]
        ]
    )

    result = _apply_symmetric_observable(
        matrix, N_qubits=1, observable=observable
    )
    assert np.allclose(result, expected)



# Copyright (C) Unitary Fund
#
# This source code is licensed under the GPL license (v3) found in the
# LICENSE file in the root directory of this source tree.

"""Tests for mitiq.pea.amplify_depolarizing functions."""

import pytest
from cirq import (
    CCNOT,
    CNOT,
    Circuit,
    H,
    LineQubit,
    MeasurementGate,
    X,
    Y,
)

from mitiq.interface import convert_from_mitiq, convert_to_mitiq
from mitiq.pea.amplifications.amplify_depolarizing import (
    amplify_noisy_op_with_global_depolarizing_noise,
    amplify_noisy_op_with_local_depolarizing_noise,
    amplify_noisy_ops_in_circuit_with_global_depolarizing_noise,
    amplify_noisy_ops_in_circuit_with_local_depolarizing_noise,
)
from mitiq.typing import SUPPORTED_PROGRAM_TYPES
from mitiq.utils import _equal


def single_qubit_depolarizing_overhead(noise_level: float) -> float:
    """See :cite:`Temme_2017_PRL` for more information.

    Args:
        noise_level: multiplier of noise level in :cite:`Temme_2017_PRL`

    Returns:
        Depolarizing overhead value with noise level considered.
    """
    epsilon = 3 / 4 * noise_level
    return 2 / 3 * (epsilon - 1)


def two_qubit_depolarizing_overhead(noise_level: float) -> float:
    """See :cite:`Temme_2017_PRL` for more information.

    Args:
        noise_level: multiplier of noise level in :cite:`Temme_2017_PRL`

    Returns:
        Depolarizing overhead value with noise level considered.
    """
    epsilon = 15 / 16 * noise_level
    return (epsilon - 1) / (epsilon + 7 / 8)


def test_three_qubit_depolarizing_amplification_error():
    q0, q1, q2 = LineQubit.range(3)
    with pytest.raises(ValueError):
        amplify_noisy_op_with_global_depolarizing_noise(
            Circuit(CCNOT(q0, q1, q2)),
            0.05,
        )


def test_three_qubit_local_depolarizing_amplification_error():
    q0, q1, q2 = LineQubit.range(3)
    with pytest.raises(ValueError):
        amplify_noisy_op_with_local_depolarizing_noise(
            Circuit(CCNOT(q0, q1, q2)),
            0.05,
        )


@pytest.mark.parametrize("circuit_type", SUPPORTED_PROGRAM_TYPES.keys())
def test_amplify_operations_in_circuit_global(circuit_type: str):
    """Tests all operation amplifications are created."""
    qreg = LineQubit.range(2)
    circ_mitiq = Circuit([CNOT(*qreg), H(qreg[0]), Y(qreg[1]), CNOT(*qreg)])
    circ = convert_from_mitiq(circ_mitiq, circuit_type)

    amps = amplify_noisy_ops_in_circuit_with_global_depolarizing_noise(
        ideal_circuit=circ,
        noise_level=0.1,
    )

    # For each operation in circ we should find its amplification
    for op in convert_to_mitiq(circ)[0].all_operations():
        found = False
        for amp in amps:
            if _equal(amp.ideal, Circuit(op), require_qubit_equality=True):
                found = True
        assert found

    # The number of amps. should match the number of unique operations
    assert len(amps) == 3


@pytest.mark.parametrize("circuit_type", SUPPORTED_PROGRAM_TYPES.keys())
def test_amplify_operations_in_circuit_local(circuit_type: str):
    """Tests all operation amplifications are created."""
    qreg = LineQubit.range(2)
    circ_mitiq = Circuit([CNOT(*qreg), H(qreg[0]), Y(qreg[1]), CNOT(*qreg)])
    circ = convert_from_mitiq(circ_mitiq, circuit_type)

    amps = amplify_noisy_ops_in_circuit_with_local_depolarizing_noise(
        ideal_circuit=circ,
        noise_level=0.1,
    )

    for op in convert_to_mitiq(circ)[0].all_operations():
        found = False
        for amp in amps:
            if _equal(amp.ideal, Circuit(op), require_qubit_equality=True):
                found = True
        assert found

    # The number of amps. should match the number of unique operations
    assert len(amps) == 3


@pytest.mark.parametrize(
    "amplification_function",
    [
        amplify_noisy_ops_in_circuit_with_local_depolarizing_noise,
        amplify_noisy_ops_in_circuit_with_global_depolarizing_noise,
    ],
)
@pytest.mark.parametrize("circuit_type", ["cirq", "qiskit", "pyquil"])
def test_amplify_operations_in_circuit_with_measurements(
    circuit_type: str,
    amplification_function,
):
    """Tests measurements in circuit are ignored (not noise amplified)."""
    q0, q1 = LineQubit.range(2)
    circ_mitiq = Circuit(
        X(q1),
        MeasurementGate(num_qubits=1)(q0),
        X(q1),
        MeasurementGate(num_qubits=1)(q0),
    )
    circ = convert_from_mitiq(circ_mitiq, circuit_type)

    amps = amplification_function(ideal_circuit=circ, noise_level=0.1)

    for op in convert_to_mitiq(circ)[0].all_operations():
        found = False
        for amp in amps:
            if _equal(amp.ideal, Circuit(op), require_qubit_equality=True):
                found = True
        if isinstance(op.gate, MeasurementGate):
            assert not found
        else:
            assert found

    # Number of unique gates excluding measurement gates
    assert len(amps) == 1



# Copyright (C) Unitary Foundation
#
# This source code is licensed under the GPL license (v3) found in the
# LICENSE file in the root directory of this source tree.

"""Tests for mitiq.pea.scale_amplifications functions."""

import pytest
from cirq import (
    CNOT,
    Circuit,
    H,
    LineQubit,
    Y,
)

from mitiq.pea.amplifications.amplify_depolarizing import (
    amplify_noisy_ops_in_circuit_with_global_depolarizing_noise,
    amplify_noisy_ops_in_circuit_with_local_depolarizing_noise,
)
from mitiq.pea.scale_amplifications import (
    scale_circuit_amplifications,
)

qreg = LineQubit.range(2)
circ = Circuit([CNOT(*qreg), H(qreg[0]), Y(qreg[1]), CNOT(*qreg)])


@pytest.mark.parametrize("epsilon", [0.01, 0.02])
@pytest.mark.parametrize(
    "noise_model, noise_function",
    [
        (
            "local_depolarizing",
            amplify_noisy_ops_in_circuit_with_local_depolarizing_noise,
        ),
        (
            "global_depolarizing",
            amplify_noisy_ops_in_circuit_with_global_depolarizing_noise,
        ),
    ],
)
def test_scale_circuit_amplifications(epsilon, noise_model, noise_function):
    scale_factors = [1, 3, 5, 7]
    amp_fn = noise_function
    for s in scale_factors:
        amps = amp_fn(circ, s * epsilon)
        scaled_amps = scale_circuit_amplifications(
            circ, s, noise_model, epsilon
        )
        assert amps == scaled_amps


def test_noise_model_not_implemented_error():
    noise_model = "amplitude_damping"
    with pytest.raises(ValueError, match="Noise model not supported"):
        scale_circuit_amplifications(circ, 1.0, noise_model, 0.01)



# Copyright (C) Unitary Foundation
#
# This source code is licensed under the GPL license (v3) found in the
# LICENSE file in the root directory of this source tree.

"""Unit tests for Collector."""

from random import choices

import cirq
import numpy as np
import pyquil
import pytest
from qiskit import QuantumCircuit

from mitiq import MeasurementResult
from mitiq.executor.executor import Executor
from mitiq.interface.mitiq_cirq import (
    compute_density_matrix,
    sample_bitstrings,
)
from mitiq.observable import Observable, PauliString


# Serial / batched executors which return floats.
def executor_batched(circuits, **kwargs) -> list[float]:
    return [
        float(v)
        for v in np.full(
            shape=(len(circuits),),
            fill_value=kwargs.setdefault("return_value", 0.0),
        )
    ]


def executor_batched_unique(circuits) -> list[float]:
    return [executor_serial_unique(circuit) for circuit in circuits]


def executor_serial_unique(circuit) -> float:
    return float(len(circuit))


def executor_serial_typed(*args, **kwargs) -> float:
    return executor_serial(*args, **kwargs)


def executor_serial(*args, **kwargs):
    return kwargs.setdefault("return_value", 0.0)


def executor_pyquil_batched(programs) -> list[float]:
    for p in programs:
        if not isinstance(p, pyquil.Program):
            raise TypeError

    return [0.0] * len(programs)


# Serial / batched executors which return measurements.
def executor_measurements(circuit):
    return sample_bitstrings(circuit, noise_level=(0,))


def executor_measurements_typed(circuit) -> MeasurementResult:
    return sample_bitstrings(circuit, noise_level=(0,))


def executor_measurements_batched(circuits) -> list[MeasurementResult]:
    return [executor_measurements_typed(circuit) for circuit in circuits]


# Serial / batched executors which return density matrices.
def executor_density_matrix(circuit):
    return compute_density_matrix(circuit, noise_level=(0,))


def executor_density_matrix_typed(circuit) -> np.ndarray:
    return compute_density_matrix(circuit, noise_level=(0,))


def executor_density_matrix_batched(circuits) -> list[np.ndarray]:
    return [executor_density_matrix_typed(circuit) for circuit in circuits]


def executor_typed_npfloat32(circuit) -> np.float32:
    return np.float32(3.14)


def test_executor_simple():
    collector = Executor(executor=executor_batched, max_batch_size=10)
    assert collector.can_batch
    assert collector._max_batch_size == 10
    assert collector.calls_to_executor == 0


def test_executors_can_batch():
    assert Executor(executor_batched).can_batch
    assert not Executor(executor_serial_typed).can_batch
    assert not Executor(executor_serial).can_batch
    assert not Executor(executor_measurements_typed).can_batch
    assert not Executor(executor_density_matrix_typed).can_batch
    assert Executor(executor_measurements_batched).can_batch


def test_executor_non_hermitian_observable():
    obs = Observable(PauliString("Z", coeff=1j))

    q = cirq.LineQubit(0)
    circuits = [cirq.Circuit(cirq.I.on(q)), cirq.Circuit(cirq.X.on(q))]

    executor = Executor(executor_measurements_typed)

    with pytest.warns(UserWarning, match="hermitian"):
        executor.evaluate(circuits, obs)


def test_run_executor_single_circuit():
    collector = Executor(executor=executor_serial)
    circuit = cirq.Circuit(cirq.H(cirq.LineQubit(0)))
    results_no_sequence = collector.run(circuit)
    results_sequence = collector.run([circuit])
    assert np.allclose(results_no_sequence, np.zeros(1))
    assert np.allclose(results_no_sequence, results_sequence)


@pytest.mark.parametrize("ncircuits", (5, 10, 25))
@pytest.mark.parametrize("executor", (executor_batched, executor_serial))
def test_run_executor_identical_circuits_batched(ncircuits, executor):
    collector = Executor(executor=executor, max_batch_size=10)
    circuits = [cirq.Circuit(cirq.H(cirq.LineQubit(0)))] * ncircuits
    results = collector.run(circuits, force_run_all=False)

    assert np.allclose(results, np.zeros(ncircuits))
    assert collector.calls_to_executor == 1


@pytest.mark.parametrize("batch_size", (1, 2, 10))
def test_run_executor_nonidentical_pyquil_programs(batch_size):
    collector = Executor(
        executor=executor_pyquil_batched, max_batch_size=batch_size
    )
    assert collector.can_batch

    circuits = [
        pyquil.Program(pyquil.gates.X(0)),
        pyquil.Program(pyquil.gates.H(0)),
    ] * 10
    results = collector.run(circuits, force_run_all=False)

    assert np.allclose(results, np.zeros(len(circuits)))
    if batch_size == 1:
        assert collector.calls_to_executor == 2
    else:
        assert collector.calls_to_executor == 1


@pytest.mark.parametrize("ncircuits", (10, 11, 23))
@pytest.mark.parametrize("batch_size", (1, 2, 5, 50))
def test_run_executor_all_unique(ncircuits, batch_size):
    collector = Executor(executor=executor_batched, max_batch_size=batch_size)
    assert collector.can_batch

    random_state = np.random.RandomState(seed=1)
    circuits = [
        cirq.testing.random_circuit(
            qubits=4, n_moments=10, op_density=1, random_state=random_state
        )
        for _ in range(ncircuits)
    ]
    results = collector.run(circuits)

    assert np.allclose(results, np.zeros(ncircuits))
    assert collector.calls_to_executor == np.ceil(ncircuits / batch_size)


@pytest.mark.parametrize("ncircuits", (5, 21))
@pytest.mark.parametrize("force_run_all", (True, False))
def test_run_executor_force_run_all_serial_executor_identical_circuits(
    ncircuits, force_run_all
):
    collector = Executor(executor=executor_serial)
    assert not collector.can_batch

    circuits = [cirq.Circuit(cirq.H(cirq.LineQubit(0)))] * ncircuits
    results = collector.run(circuits, force_run_all=force_run_all)

    assert np.allclose(results, np.zeros(ncircuits))
    if force_run_all:
        assert collector.calls_to_executor == ncircuits
    else:
        assert collector.calls_to_executor == 1


@pytest.mark.parametrize("s", (50, 100, 150))
@pytest.mark.parametrize("b", (1, 2, 100))
def test_run_executor_preserves_order(s, b):
    collector = Executor(executor=executor_batched_unique, max_batch_size=b)
    assert collector.can_batch

    circuits = [
        cirq.Circuit(cirq.H(cirq.LineQubit(0))),
        cirq.Circuit([cirq.H(cirq.LineQubit(0))] * 2),
    ]
    batch = choices(circuits, k=s)

    assert np.allclose(collector.run(batch), executor_batched_unique(batch))


@pytest.mark.parametrize(
    "execute", [executor_serial_unique, executor_batched_unique]
)
def test_executor_evaluate_float(execute):
    q = cirq.LineQubit(0)
    circuits = [
        cirq.Circuit(cirq.X(q), cirq.M(q)),
        cirq.Circuit(cirq.H(q), cirq.Z(q), cirq.M(q)),
    ]

    executor = Executor(execute)

    executor.evaluate(circuits)

    if execute is executor_serial_unique:
        assert executor.calls_to_executor == 2
    else:
        assert executor.calls_to_executor == 1

    assert executor.executed_circuits == circuits
    assert executor.quantum_results == [2, 3]


@pytest.mark.parametrize(
    "execute", [executor_measurements_typed, executor_measurements_batched]
)
def test_executor_evaluate_measurements(execute):
    obs = Observable(PauliString("Z"))

    q = cirq.LineQubit(0)
    circuits = [cirq.Circuit(cirq.I.on(q)), cirq.Circuit(cirq.X.on(q))]

    executor = Executor(execute)

    results = executor.evaluate(circuits, obs)
    assert np.allclose(results, [1, -1])

    if execute is executor_measurements_typed:
        assert executor.calls_to_executor == 2
    else:
        assert executor.calls_to_executor == 1

    assert executor.executed_circuits[0] == circuits[0] + cirq.measure(q)
    assert executor.executed_circuits[1] == circuits[1] + cirq.measure(q)
    assert executor.quantum_results[0] == executor_measurements_typed(
        circuits[0] + cirq.measure(q)
    )
    assert executor.quantum_results[1] == executor_measurements_typed(
        circuits[1] + cirq.measure(q)
    )
    assert len(executor.quantum_results) == len(circuits)


@pytest.mark.parametrize(
    "execute", [executor_density_matrix_typed, executor_density_matrix_batched]
)
def test_executor_evaluate_density_matrix(execute):
    obs = Observable(PauliString("Z"))

    q = cirq.LineQubit(0)
    circuits = [cirq.Circuit(cirq.I.on(q)), cirq.Circuit(cirq.X.on(q))]

    executor = Executor(execute)

    results = executor.evaluate(circuits, obs)
    assert np.allclose(results, [1, -1])

    if execute is executor_density_matrix_typed:
        assert executor.calls_to_executor == 2
    else:
        assert executor.calls_to_executor == 1

    assert executor.executed_circuits == circuits
    assert np.allclose(
        executor.quantum_results[0], executor_density_matrix_typed(circuits[0])
    )
    assert np.allclose(
        executor.quantum_results[1], executor_density_matrix_typed(circuits[1])
    )
    assert len(executor.quantum_results) == len(circuits)


def test_executor_float_with_observable_typed():
    obs = Observable(PauliString("Z"))
    q = cirq.LineQubit(0)
    circuit = cirq.Circuit(cirq.X.on(q))
    executor = Executor(executor_serial_typed)
    with pytest.raises(
        ValueError,
        match="When using an executor which returns a float-like ",
    ):
        executor.evaluate(circuit, obs)


def test_executor_measurements_without_observable_typed():
    q = cirq.LineQubit(0)
    circuit = cirq.Circuit(cirq.X.on(q))
    executor = Executor(executor_measurements_typed)
    with pytest.raises(
        ValueError,
        match="When using a measurement, or bitstring, like result",
    ):
        executor.evaluate(circuit)


def test_executor_density_matrix_without_observable_typed():
    q = cirq.LineQubit(0)
    circuit = cirq.Circuit(cirq.X.on(q))
    executor = Executor(executor_density_matrix_typed)
    with pytest.raises(
        ValueError,
        match="When using a density matrix result",
    ):
        executor.evaluate(circuit)


def test_executor_float_not_typed():
    executor = Executor(executor_serial)
    executor_typed = Executor(executor_serial_typed)
    qcirc = QuantumCircuit(1)
    qcirc.h(0)
    assert executor.evaluate(qcirc) == executor_typed.evaluate(qcirc)


def test_executor_density_matrix_not_typed():
    obs = Observable(PauliString("Z"))
    executor = Executor(executor_density_matrix)
    q = cirq.LineQubit(0)
    circuit = cirq.Circuit(cirq.X.on(q))
    with pytest.raises(
        ValueError,
        match="When using an observable",
    ):
        executor.evaluate(circuit, obs)


def test_executor_measurements_not_typed():
    obs = Observable(PauliString("Z"))
    executor = Executor(executor_measurements)
    q = cirq.LineQubit(0)
    circuit = cirq.Circuit(cirq.X.on(q))
    with pytest.raises(
        ValueError,
        match="When using an observable",
    ):
        executor.evaluate(circuit, obs)


def test_executor_unknown_type():
    obs = Observable(PauliString("Z"))
    executor = Executor(executor_typed_npfloat32)
    q = cirq.LineQubit(0)
    circuit = cirq.Circuit(cirq.X.on(q))
    with pytest.raises(
        ValueError,
        match="Could not parse executed results",
    ):
        executor.evaluate(circuit, obs)



# Copyright (C) Unitary Foundation
#
# This source code is licensed under the GPL license (v3) found in the
# LICENSE file in the root directory of this source tree.

"""Run experiments without error mitigation."""

from collections.abc import Callable

from mitiq import QPROGRAM, Executor, Observable, QuantumResult


def execute(
    circuit: QPROGRAM,
    executor: Executor | Callable[[QPROGRAM], QuantumResult],
    observable: Observable | None = None,
) -> float:
    """Evaluates the expectation value associated to the input circuit
    without using error mitigation.

    The only purpose of this function is to provide the same interface for
    non-error-mitigated values as the rest of the techniques in Mitiq. This
    is useful when comparing error-mitigated results to non-error-mitigated
    results.

    Args:
        circuit: The circuit to run.
        executor: Executes a circuit and returns a `QuantumResult`.
        observable: Observable to compute the expectation value of. If None,
            the `executor` must return an expectation value. Otherwise,
            the `QuantumResult` returned by `executor` is used to compute the
            expectation of the observable.
    """
    if not isinstance(executor, Executor):
        executor = Executor(executor)

    return executor.evaluate(circuit, observable)[0]



# Copyright (C) Unitary Foundation
#
# This source code is licensed under the GPL license (v3) found in the
# LICENSE file in the root directory of this source tree.

"""Tools for constructing the noise-amplified representations of ideal
operations.
"""

from collections.abc import Sequence

from cirq import Circuit

from mitiq.pea.amplifications.amplify_depolarizing import (
    amplify_noisy_ops_in_circuit_with_global_depolarizing_noise,
    amplify_noisy_ops_in_circuit_with_local_depolarizing_noise,
)
from mitiq.pec import OperationRepresentation


def scale_circuit_amplifications(
    ideal_circuit: Circuit,
    scale_factor: float,
    noise_model: str,
    epsilon: float,
) -> Sequence[OperationRepresentation]:
    r"""Generates a list of implementable sequences from the noise-amplified
    representation of the input ideal circuit based on the input noise model
    and baseline noise level.

    Args:
        ideal_circuit: The ideal circuit from which an implementable
            sequence is sampled.
        scale_factor: A (positive) number by which the baseline noise
            level is to be amplified.
        noise_model: A string describing the noise model to be used for the
            noise-scaled representations, e.g. "local_depolarizing" or
            "global_depolarizing".
        epsilon: Baseline noise level.

    Returns:
        A list of noise-amplified circuits, corresponding to each scale
        factor multiplied by the baseline noise level."""

    if noise_model == "local_depolarizing":
        amp_fn = amplify_noisy_ops_in_circuit_with_local_depolarizing_noise
        # TODO add other existing noise models from Mitiq
    elif noise_model == "global_depolarizing":
        amp_fn = amplify_noisy_ops_in_circuit_with_global_depolarizing_noise
    else:
        raise ValueError("Noise model not supported")
        # TODO allow use of custom noise model

    return amp_fn(ideal_circuit, scale_factor * epsilon)



# Copyright (C) Unitary Foundation
#
# This source code is licensed under the GPL license (v3) found in the
# LICENSE file in the root directory of this source tree.
"""Tests for Cirq executors defined in cirq_utils.py"""

import cirq
import numpy as np

from mitiq import MeasurementResult
from mitiq.interface.mitiq_cirq import (
    compute_density_matrix,
    execute_with_depolarizing_noise,
    sample_bitstrings,
)


def test_sample_bitstrings():
    """Tests if the outcome is as expected with different noise models and
    error rate."""

    # testing default options i.e. noise model is amplitude damping with
    # the default error rate, shots is also the default number
    qc = cirq.Circuit(cirq.X(cirq.LineQubit(0))) + cirq.Circuit(
        cirq.measure(cirq.LineQubit(0))
    )
    result_default = sample_bitstrings(qc)
    assert result_default.nqubits == 1
    assert result_default.qubit_indices == (0,)
    assert result_default.shots == 8192
    assert isinstance(result_default, MeasurementResult)

    # test for sum(noise_level) = 0
    result_no_noise = sample_bitstrings(qc, noise_level=(0.00,), shots=10)
    assert np.allclose(
        result_no_noise.asarray,
        np.array([[1], [1], [1], [1], [1], [1], [1], [1], [1], [1]]),
    )

    # test result with inputs different from default
    result_not_default = sample_bitstrings(
        qc,
        cirq.GeneralizedAmplitudeDampingChannel,
        (0.1, 0.1),
        cirq.DensityMatrixSimulator(),
        8000,
    )
    assert result_not_default.nqubits == 1
    assert result_not_default.qubit_indices == (0,)
    assert result_not_default.shots == 8000


def test_compute_density_matrix():
    """Tests if the density matrix of a noisy circuit is output as expected."""

    qc = cirq.Circuit(cirq.X(cirq.LineQubit(0)))
    assert np.isclose(np.trace(compute_density_matrix(qc)), 1)
    assert np.isclose(
        np.trace(
            compute_density_matrix(
                qc, cirq.GeneralizedAmplitudeDampingChannel, (0.1, 0.1)
            )
        ),
        1,
    )


def test_execute_with_depolarizing_noise():
    """Tests if executor function for Cirq returns a proper expectation
    value when used for noisy depoalrizing simulation."""

    qc = cirq.Circuit()
    for _ in range(100):
        qc += cirq.X(cirq.LineQubit(0))

    observable = np.diag([0, 1])
    # Test 1
    noise1 = 0.0
    observable_exp_value = execute_with_depolarizing_noise(
        qc, observable, noise1
    )
    assert 0.0 == observable_exp_value

    # Test 2
    noise2 = 0.5
    observable_exp_value = execute_with_depolarizing_noise(
        qc, observable, noise2
    )
    assert np.isclose(observable_exp_value, 0.5)

    # Test 3
    noise3 = 0.001
    observable_exp_value = execute_with_depolarizing_noise(
        qc, observable, noise3
    )
    assert np.isclose(observable_exp_value, 0.062452)



# Copyright (C) Unitary Foundation
#
# This source code is licensed under the GPL license (v3) found in the
# LICENSE file in the root directory of this source tree.

"""High-level probabilistic error cancellation tools."""

import warnings
from collections.abc import Callable, Iterable, Sequence
from functools import wraps
from typing import Any, cast

import numpy as np

from mitiq import QPROGRAM, Executor, Observable, QuantumResult
from mitiq.pec import OperationRepresentation, sample_circuit


class LargeSampleWarning(Warning):
    """Warning is raised when PEC sample size is greater than 10 ** 5"""

    pass


_LARGE_SAMPLE_WARN = (
    "The number of PEC samples is very large. It may take several minutes."
    " It may be necessary to reduce 'precision' or 'num_samples'."
)


def construct_circuits(
    circuit: QPROGRAM,
    representations: Sequence[OperationRepresentation],
    precision: float = 0.03,
    num_samples: int | None = None,
    random_state: int | np.random.RandomState | None = None,
    full_output: bool = False,
) -> list[QPROGRAM] | tuple[list[QPROGRAM], list[int], float]:
    """Generates a list of sampled circuits based on the given
    quasi-probability representations.

    Args:
        circuit: The quantum circuit to be sampled.
        representations: The quasi-probability representations of the circuit
            operations.
        precision: The desired precision for the sampling process.
            Default is 0.03.
        num_samples: The number of samples to generate. If None, the number of
            samples is deduced based on the precision. Default is None.
        random_state: The random state or seed for reproducibility.
        full_output: If True, returns the signs and the norm along with the
            sampled circuits. Default is False.

    Returns:
        A list of sampled circuits. If ``full_output`` is True, also returns a
        list of signs, the norm.

    Raises:
        ValueError: If the precision is not within the interval (0, 1].
    """
    if isinstance(random_state, int):
        random_state = np.random.RandomState(random_state)

    if not (0 < precision <= 1):
        raise ValueError(
            "The value of 'precision' should be within the interval (0, 1],"
            f" but precision is {precision}."
        )

    # Get the 1-norm of the circuit quasi-probability representation
    _, _, norm = sample_circuit(
        circuit,
        representations,
        num_samples=1,
    )

    # Deduce the number of samples (if not given by the user)
    if num_samples is None:
        num_samples = int((norm / precision) ** 2)

    if num_samples > 10**5:
        warnings.warn(_LARGE_SAMPLE_WARN, LargeSampleWarning)

    sampled_circuits, signs, _ = sample_circuit(
        circuit,
        representations,
        random_state=random_state,
        num_samples=num_samples,
    )

    if full_output:
        return sampled_circuits, signs, norm
    return sampled_circuits


def combine_results(
    results: Iterable[float], norm: float, signs: Iterable[int]
) -> float:
    """Combine expectation values coming from probabilistically sampled
    circuits.

    Warning:
        The ``results`` must be in the same order as the circuits were
        generated.

    Args:
        results: Results as obtained from running circuits.
        norm: The one-norm of the circuit representation.
        signs: The signs corresponding to the positivity of the sampled
            circuits.

    Returns:
        The PEC estimate of the expectation value.
    """
    unbiased_estimators = [norm * s * val for s, val in zip(signs, results)]

    pec_value = cast(float, np.average(unbiased_estimators))
    return pec_value


def execute_with_pec(
    circuit: QPROGRAM,
    executor: Executor | Callable[[QPROGRAM], QuantumResult],
    observable: Observable | None = None,
    *,
    representations: Sequence[OperationRepresentation],
    precision: float = 0.03,
    num_samples: int | None = None,
    force_run_all: bool = True,
    random_state: int | np.random.RandomState | None = None,
    full_output: bool = False,
) -> float | tuple[float, dict[str, Any]]:
    r"""Estimates the error-mitigated expectation value associated to the
    input circuit, via the application of probabilistic error cancellation
    (PEC). :cite:`Temme_2017_PRL` :cite:`Endo_2018_PRX`.

    This function implements PEC by:

    1. Sampling different implementable circuits from the quasi-probability
       representation of the input circuit;
    2. Evaluating the noisy expectation values associated to the sampled
       circuits (through the "executor" function provided by the user);
    3. Estimating the ideal expectation value from a suitable linear
       combination of the noisy ones.

    Args:
        circuit: The input circuit to execute with error-mitigation.
        executor: A Mitiq executor that executes a circuit and returns the
            unmitigated ``QuantumResult`` (e.g. an expectation value).
        observable: Observable to compute the expectation value of. If None,
            the `executor` must return an expectation value. Otherwise,
            the `QuantumResult` returned by `executor` is used to compute the
            expectation of the observable.
        representations: Representations (basis expansions) of each operation
            in the input circuit.
        precision: The desired estimation precision (assuming the observable
            is bounded by 1). The number of samples is deduced according
            to the formula (one_norm / precision) ** 2, where 'one_norm'
            is related to the negativity of the quasi-probability
            representation :cite:`Temme_2017_PRL`. If 'num_samples' is
            explicitly set by the user, 'precision' is ignored and has no
            effect.
        num_samples: The number of noisy circuits to be sampled for PEC.
            If not given, this is deduced from the argument 'precision'.
        force_run_all: If True, all sampled circuits are executed regardless of
            uniqueness, else a minimal unique set is executed.
        random_state: Seed for sampling circuits.
        full_output: If False only the average PEC value is returned.
            If True a dictionary containing all PEC data is returned too.

    Returns:
        The tuple ``(pec_value, pec_data)`` where ``pec_value`` is the
        expectation value estimated with PEC and ``pec_data`` is a dictionary
        which contains all the raw data involved in the PEC process (including
        the PEC estimation error).
        The error is estimated as ``pec_std / sqrt(num_samples)``, where
        ``pec_std`` is the standard deviation of the PEC samples, i.e., the
        square root of the mean squared deviation of the sampled values from
        ``pec_value``. If ``full_output`` is ``False``, only ``pec_value`` is
        returned.
    """
    sampled_circuits, signs, norm = construct_circuits(
        circuit,
        representations,
        precision,
        num_samples,
        random_state=random_state,
        full_output=True,
    )

    # Execute all sampled circuits
    if not isinstance(executor, Executor):
        executor = Executor(executor)

    results = executor.evaluate(sampled_circuits, observable, force_run_all)

    # Evaluate unbiased estimators [Temme2017] [Endo2018] [Takagi2020]
    unbiased_estimators = [norm * s * val for s, val in zip(signs, results)]

    pec_value = cast(float, np.average(unbiased_estimators))

    if not full_output:
        return pec_value

    num_circuits = len(sampled_circuits)
    # Build dictionary with additional results and data
    pec_data: dict[str, Any] = {
        "num_samples": num_circuits,
        "precision": precision,
        "pec_value": pec_value,
        "pec_error": np.std(unbiased_estimators) / np.sqrt(num_circuits),
        "unbiased_estimators": unbiased_estimators,
        "measured_expectation_values": results,
        "sampled_circuits": sampled_circuits,
    }

    return pec_value, pec_data


def mitigate_executor(
    executor: Callable[[QPROGRAM], QuantumResult],
    observable: Observable | None = None,
    *,
    representations: Sequence[OperationRepresentation],
    precision: float = 0.03,
    num_samples: int | None = None,
    force_run_all: bool = True,
    random_state: int | np.random.RandomState | None = None,
    full_output: bool = False,
) -> Callable[[QPROGRAM], float | tuple[float, dict[str, Any]]]:
    """Returns a modified version of the input 'executor' which is
    error-mitigated with probabilistic error cancellation (PEC).

    Args:
        executor: A function that executes a circuit and returns the
            unmitigated `QuantumResult` (e.g. an expectation value).
        observable: Observable to compute the expectation value of. If None,
            the `executor` must return an expectation value. Otherwise,
            the `QuantumResult` returned by `executor` is used to compute the
            expectation of the observable.
        representations: Representations (basis expansions) of each operation
            in the input circuit.
        precision: The desired estimation precision (assuming the observable
            is bounded by 1). The number of samples is deduced according
            to the formula (one_norm / precision) ** 2, where 'one_norm'
            is related to the negativity of the quasi-probability
            representation :cite:`Temme_2017_PRL`. If 'num_samples' is
            explicitly set, 'precision' is ignored and has no effect.
        num_samples: The number of noisy circuits to be sampled for PEC.
            If not given, this is deduced from the argument 'precision'.
        force_run_all: If True, all sampled circuits are executed regardless of
            uniqueness, else a minimal unique set is executed.
        random_state: Seed for sampling circuits.
        full_output: If False only the average PEC value is returned.
            If True a dictionary containing all PEC data is returned too.

    Returns:
        The error-mitigated version of the input executor.
    """
    executor_obj = Executor(executor)
    if not executor_obj.can_batch:

        @wraps(executor)
        def new_executor(
            circuit: QPROGRAM,
        ) -> float | tuple[float, dict[str, Any]]:
            return execute_with_pec(
                circuit,
                executor,
                observable,
                representations=representations,
                precision=precision,
                num_samples=num_samples,
                force_run_all=force_run_all,
                random_state=random_state,
                full_output=full_output,
            )

    else:

        @wraps(executor)
        def new_executor(
            circuits: list[QPROGRAM],
        ) -> list[float | tuple[float, dict[str, Any]]]:
            return [
                execute_with_pec(
                    circuit,
                    executor,
                    observable,
                    representations=representations,
                    precision=precision,
                    num_samples=num_samples,
                    force_run_all=force_run_all,
                    random_state=random_state,
                    full_output=full_output,
                )
                for circuit in circuits
            ]

    return new_executor


def pec_decorator(
    observable: Observable | None = None,
    *,
    representations: Sequence[OperationRepresentation],
    precision: float = 0.03,
    num_samples: int | None = None,
    force_run_all: bool = True,
    random_state: int | np.random.RandomState | None = None,
    full_output: bool = False,
) -> Callable[
    [Callable[[QPROGRAM], QuantumResult]],
    Callable[
        [QPROGRAM],
        float | tuple[float, dict[str, Any]],
    ],
]:
    """Decorator which adds an error-mitigation layer based on probabilistic
    error cancellation (PEC) to an executor function, i.e., a function which
    executes a quantum circuit with an arbitrary backend and returns a
    ``QuantumResult`` (e.g. an expectation value).


    Args:
        observable: Observable to compute the expectation value of. If None,
            the `executor` function being decorated must return an expectation
            value. Otherwise, the `QuantumResult` returned by this `executor`
            is used to compute the expectation of the observable.
        representations: Representations (basis expansions) of each operation
            in the input circuit.
        precision: The desired estimation precision (assuming the observable
            is bounded by 1). The number of samples is deduced according
            to the formula (one_norm / precision) ** 2, where 'one_norm'
            is related to the negativity of the quasi-probability
            representation :cite:`Temme_2017_PRL`. If 'num_samples' is
            explicitly set by the user, 'precision' is ignored and has no
            effect.
        num_samples: The number of noisy circuits to be sampled for PEC.
            If not given, this is deduced from the argument 'precision'.
        force_run_all: If True, all sampled circuits are executed regardless of
            uniqueness, else a minimal unique set is executed.
        random_state: Seed for sampling circuits.
        full_output: If False only the average PEC value is returned.
            If True a dictionary containing all PEC data is returned too.

    Returns:
        The error-mitigating decorator to be applied to an executor function.
    """

    def decorator(
        executor: Callable[[QPROGRAM], QuantumResult],
    ) -> Callable[[QPROGRAM], float | tuple[float, dict[str, Any]]]:
        return mitigate_executor(
            executor,
            observable,
            representations=representations,
            precision=precision,
            num_samples=num_samples,
            force_run_all=force_run_all,
            random_state=random_state,
            full_output=full_output,
        )

    return decorator
